package rendering

import (
	"context"
	"errors"
	"fmt"
	"sort"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/utils"
)

var (
	// ErrDuplicateName is returned by EngineRegistry.Register if and engine with
	// the same name is already registered.
	ErrDuplicateName = errors.New("duplicate engine name")

	ErrEmptyPipeline = errors.New("empty pipeline")

	ErrNoEngine = errors.New("no engine found to convert")
)

// EngineTransitionInfo describes a.
type EngineTransitionInfo struct {
	FromType string
	ToType   string
	Priority float64
}

// EngineMetadataInfo ...
type EngineMetadataInfo struct {
	Name        string
	Default     string
	Description string
}

// EngineInfo describes an engine capabilities.
type EngineInfo struct {
	Type        string
	Description string
	Transitions []EngineTransitionInfo
	Metadata    []EngineMetadataInfo
}

// Engine is able to render a document.
type Engine interface {
	Info() EngineInfo
	Render(ctx context.Context, w DocumentWriter, r DocumentReader, toType string) error
}

// RegisterEngine a new engine factory.
func RegisterEngine(name string, engine Engine) error {
	return DefaultEngineRegistry.Register(name, engine)
}

func RegisterPooledEngine(
	name string,
	factory func(zerolog.Logger) (Engine, error),
	size int,
	log zerolog.Logger,
) error {
	return DefaultEngineRegistry.RegisterPooled(name, factory, size, log)
}

// MustRegisterEngine registers an Engine and panics if it fails.
func MustRegisterEngine(name string, engine Engine) {
	DefaultEngineRegistry.MustRegister(name, engine)
}

func MustRegisterPooledEngine(name string, factory func(zerolog.Logger) (Engine, error), size int, log zerolog.Logger) {
	DefaultEngineRegistry.MustRegisterPooled(name, factory, size, log)
}

// EngineRegistry is a set of Engine, ready for use.
type EngineRegistry struct {
	Engines map[string]Engine
}

// Register a Engine.
func (s *EngineRegistry) Register(name string, engine Engine) error {
	if _, ok := s.Engines[name]; ok {
		return ErrDuplicateName
	}

	s.Engines[name] = engine

	return nil
}

func (s *EngineRegistry) RegisterPooled(
	name string,
	factory func(zerolog.Logger) (Engine, error),
	size int,
	log zerolog.Logger,
) error {
	pooledEngine, err := NewPooledEngine(
		factory, size, log.With().Str("engine-name", name).Logger())
	if err != nil {
		return err
	}

	return s.Register(name, pooledEngine)
}

// MustRegister an Engine.
func (s *EngineRegistry) MustRegister(name string, engine Engine) {
	if err := s.Register(name, engine); err != nil {
		panic(err)
	}
}

// MustRegisterPooled an Engine.
func (s *EngineRegistry) MustRegisterPooled(
	name string,
	factory func(zerolog.Logger) (Engine, error),
	size int,
	log zerolog.Logger,
) {
	if err := s.RegisterPooled(name, factory, size, log); err != nil {
		panic(err)
	}
}

// PipelineStep is a step in a pipeline.
type PipelineStep struct {
	SourceType string
	TargetType string
	Engine     Engine
}

type pipelineStepResult struct {
	step PipelineStep
	err  error
}

type Pipeline []PipelineStep

func (p Pipeline) Render(ctx context.Context, w DocumentWriter, r DocumentReader) error {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	if len(p) == 0 {
		return fmt.Errorf("cannot render: %w", ErrEmptyPipeline)
	}

	results := make(chan pipelineStepResult, len(p))

	runStep := func(step PipelineStep, w DocumentWriter, r DocumentReader) {
		timer := utils.NewVecTimer(renderDocumentDuration)

		w.SetType(step.TargetType)
		result := pipelineStepResult{
			step: step,
			err:  step.Engine.Render(ctx, w, r, step.TargetType),
		}
		status := "success"

		if result.err != nil {
			status = "failure"
		}

		if err := w.Close(); err != nil {
			if result.err == nil {
				result.err = err
				status = "failure"
			}
		}

		timer.ObserveDurationWithLabelValues(
			step.SourceType, step.TargetType, step.Engine.Info().Type, status)
		results <- result
	}

	for _, step := range p[:len(p)-1] {
		pw, pr := DocumentPipe()
		go runStep(step, pw, r)
		r = pr
	}

	go runStep(p[len(p)-1], w, r)

	var err error

	for i := 0; i != len(p); i++ {
		stepResult := <-results
		if stepResult.err != nil {
			if err == nil {
				err = fmt.Errorf(
					"engine %s failed: %w",
					stepResult.step.Engine.Info().Type, stepResult.err)

				cancel()
			}
		}
	}

	return err
}

var paths = [][]string{
	{
		"text/mjml",
		"text/html",
		"application/pdf",
	},
	{
		"text/mjml",
		"text/html",
		"application/pdf",
		"image/png",
	},
	{
		"text/plain",
		"application/pdf",
		"image/png",
	},
	{
		"application/vnd.oasis.opendocument.text",
		"application/pdf",
		"image/png",
	},
	{
		"application/msword",
		"application/pdf",
		"image/png",
	},
	{
		"application/vnd.openxmlformats-officedocument.wordprocessingml.document",
		"application/pdf",
		"image/png",
	},
	{
		"text/html",
		"application/pdf",
		"image/png",
	},
}

// GetPipeline returns a list of engine for converting a document of type 'fromType'
// to type 'toType'
// TODO do a real implementation, not this hard-coded hack.
func (s *EngineRegistry) GetPipeline(fromType, toType string) (Pipeline, error) {
	if fromType == toType {
		return nil, nil
	}

	for _, p := range paths {
		if p[0] == fromType && p[len(p)-1] == toType {
			pipeline := make([]PipelineStep, len(p)-1)

			from := p[0]
			for i, to := range p[1:] {
				engine, err := s.FindEngine(from, to)
				if err != nil {
					return nil, err
				}

				pipeline[i] = PipelineStep{from, to, engine}
				from = to
			}

			return pipeline, nil
		}
	}

	engine, err := s.FindEngine(fromType, toType)
	if err != nil {
		return nil, err
	}

	return Pipeline{
		{fromType, toType, engine},
	}, nil
}

// FindEngine finds the engine to do a transformation that has the highest priority.
func (s *EngineRegistry) FindEngine(fromType, toType string) (Engine, error) {
	engines, err := s.FindEngines(fromType, toType)
	if err != nil {
		return nil, err
	}

	if len(engines) == 0 {
		return nil, fmt.Errorf("%w from %s to %s", ErrNoEngine, fromType, toType)
	}

	return engines[0], nil
}

// FindEngines finds engines that can do a given transformation. The result
// is sorted by priority, the highest first.
func (s *EngineRegistry) FindEngines(fromType, toType string) ([]Engine, error) {
	type entry struct {
		engine   Engine
		priority float64
	}

	entries := []entry{}

	for _, engine := range s.Engines {
		for _, tr := range engine.Info().Transitions {
			if tr.FromType == fromType && tr.ToType == toType {
				entries = append(entries, entry{engine, tr.Priority})
			}
		}
	}

	sort.Slice(entries, func(i, j int) bool {
		return entries[i].priority > entries[j].priority
	})

	engines := make([]Engine, 0, len(entries))
	for _, entry := range entries {
		engines = append(engines, entry.engine)
	}

	return engines, nil
}

func (s *EngineRegistry) Close() error {
	var errs utils.GroupedError

	for _, e := range s.Engines {
		if p, ok := e.(*PooledEngine); ok {
			p.Close()
		}

		if c, ok := e.(Closer); ok {
			if err := c.Close(); err != nil {
				errs = append(errs, err)
			}
		}
	}

	s.Engines = make(map[string]Engine)

	if !errs.IsEmpty() {
		return fmt.Errorf("engines: %w", errs)
	}

	return nil
}

// NewEngineRegistry creates a EngineRegistry.
func NewEngineRegistry() *EngineRegistry {
	return &EngineRegistry{
		Engines: make(map[string]Engine),
	}
}

// DefaultEngineRegistry is the default engine registry.
var (
	DefaultEngineRegistry = NewEngineRegistry()

	renderDocumentDuration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "rednerd_render_document",
			Help:       "Render document duration and count",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
		},
		[]string{
			"from_type",
			"to_type",
			"engine",
			"status",
		},
	)
)

func init() {
	prometheus.MustRegister(renderDocumentDuration)
}
