package handlers

import (
	"context"
	"database/sql"
	"errors"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	op_auth "orus.io/orus-io/rednerd/restapi/operations/auth"
)

func NewResetPassword(db *sqlx.DB) *ResetPassword {
	return &ResetPassword{
		db: db,
	}
}

type ResetPassword struct {
	db *sqlx.DB
}

func (h *ResetPassword) handle(ctx context.Context, token, password string) error {
	log := zerolog.Ctx(ctx)

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return err
	}

	defer tx.RollbackIfOpened(*log)

	db := database.NewSQLHelper(ctx, tx, *log)

	var resetPassword models.ResetPasswordTokenDB

	if err := db.GetBy(&resetPassword, "token", token); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return &models.Error{Message: "invalid token"}
		}

		return err
	}

	var userDB models.UserDB

	if err := db.GetByPKey(&userDB, resetPassword.Username); err != nil {
		return err
	}

	if err := userDB.SetPassword(password); err != nil {
		return err
	}

	if err := db.Update(userDB); err != nil {
		return err
	}

	if _, err := db.Exec(
		squirrel.Delete(models.ResetPasswordTokenDBTable).
			Where(squirrel.Eq{models.ResetPasswordTokenDBTokenColumn: token}),
	); err != nil {
		return err
	}

	return tx.Commit()
}

func (h *ResetPassword) Handle(params op_auth.ResetPasswordParams) middleware.Responder {
	ctx := params.HTTPRequest.Context()

	if params.Newpassword.Token == "" {
		return op_auth.NewResetPasswordDefault(400).
			WithPayload(&models.Error{Message: "'token' is required"})
	}

	if err := h.handle(ctx, params.Newpassword.Token, params.Newpassword.Password); err != nil {
		var mErr *models.Error
		if errors.As(err, &mErr) {
			return op_auth.NewResetPasswordDefault(400).
				WithPayload(mErr)
		}

		return InternalError(err, zerolog.Ctx(ctx), false)
	}

	return op_auth.NewResetPasswordNoContent()
}
